float val[][] = new float[256][256];
float valPrev[][] = new float[256][256];
float delx = 1;
float delxSq = delx*delx;
float delt = 4;
float diffusionRate = 1.0;
float r = diffusionRate*delt/delxSq;

float boundStart = 0;
float boundEnd = 1;
float t = 0;
int[] sources = new int[100];
int[] sinks = new int[100];
int dbm_count = 1;
int[] dbm = new int[val.length*val.length/2];
int neighbor_count = 0;
int[] neighbors = new int[val.length*val.length/2];

float totalBound = 0;

void setup() {
  size(512,300);
  println(r);
  for(int i=0;i<val.length;++i) {
    for(int j=0;j<val[0].length;++j) {
      //val[i][j] = sq(cos(PI*(i*i+j*j)/(2.0*sq(val.length))));
      val[i][j] = 0;
      valPrev[i][j] = val[i][j];
    }
  }
  dbm[0] = 128*val.length+128;
  val[128][128] = valPrev[128][128] = -1;
  for(int i=0;i<3000;++i) stepDufortFrankel();
}

void draw() {
  background(255);
  //draw colors
  loadPixels();
  for(int i=0;i<val.length;++i) {
    for(int j=0;j<val[0].length;++j) {
      pixels[j*width+i] = color(50,constrain(val[i][j],0,1)*200,20);
    }
  }
  for(int i=0;i<dbm_count;++i) {
    int xind = dbm[i]%val.length;
    int yind = dbm[i]/val.length;
    pixels[yind*width+xind] = color(255);
    pixels[yind*width+xind+1] = color(255);
    pixels[(yind+1)*width+xind] = color(255);
    pixels[(yind+1)*width+xind+1] = color(255);
  }
  updatePixels();
  for(int i=0;i<300;++i) stepDufortFrankel();
  updateDBM();
}

//different boundary conditions

void stepPeriodic() {
  //periodic boundary conditions
  float valNext[][] = new float[256][256];
  for(int i=0;i<val.length;++i) {
    for(int j=0;j<val[0].length;++j) {
      int iPrev = (i-1+val.length)%val.length;
      int iNext = (i+1)%val.length;
      int jPrev = (j-1+val[0].length)%val[0].length;
      int jNext = (j+1)%val[0].length;
      valNext[i][j] = val[i][j]+delt*diffusionRate*(val[iPrev][j]+val[iNext][j]+val[i][jPrev]+val[i][jNext]-4*val[i][j])/delxSq;
    }
  }
  val = valNext;
}

void stepDufortFrankel() {
  totalBound = 0;
  neighbor_count = 0;
  float valNext[][] = new float[256][256];
  for(int i=0;i<val.length;++i) {
    for(int j=0;j<val[0].length;++j) {
      if(val[i][j] == -1) {
        valNext[i][j] = -1;
      } else {
        float vPrev, vNext, jNext, jPrev;
        if(i==0) vPrev = 1;//val[i][j];
        else vPrev = val[i-1][j];
        if(i==val.length-1) vNext = 1;//val[i][j];
        else vNext = val[i+1][j];
        if(j==0) jPrev = 1; //val[i][j];
        else jPrev = val[i][j-1];
        if(j==val[0].length-1) jNext = 1;
        else jNext = val[i][j+1];
        valNext[i][j] = (1-2*r)/(1+2*r)*(valPrev[i][j])+r/(1+2*r)*(max(vPrev,0)+max(vNext,0)+max(jPrev,0)+max(jNext,0));
        if(jNext == -1 || jPrev == -1 || vNext == -1 || vPrev == -1) {
          totalBound += valNext[i][j];
          neighbors[neighbor_count++] = j*val.length+i;
        }
      }
    }
  }
  valPrev = val;
  val = valNext;
  for(int i=0;i<dbm_count;++i) {
    val[dbm[i]%val.length][dbm[i]/val.length] = -1;
  }
}

void updateDBM() {
  for(int i=0;i<neighbor_count;++i) {
    int xind = neighbors[i]%val.length;
    int yind = neighbors[i]/val.length;
    if(random(1) < val[xind][yind]/totalBound) {
      dbm[dbm_count++] = neighbors[i];
      val[xind][yind] = -1;
    }
  }
}
