import controlP5.*;

int w = 128;
int h = 128;

float a[][] = new float[w][h];
float b[][] = new float[w][h];

float delx = 1; //1
float delxSq = delx*delx;

float deltaT = 5;

//REACTION VARIABLES
float reactionRate = 1.0;//0 //1.0/128.0; //Turing 1/128.0
float aRate = .1;//.1 //.2 //1/16:1/14 OR 1/16:1
float bRate = .05;//.05//.1 //.25

float decayA = 12.0; //Turing 12.0
float decayB = 16.0; //Turing 16.0

float F = .03; //.04
float k = .055; //.06

float gamma = .0004;

float maxA = 0;
float minA = 999;
float maxB = 0;
float minB = 999;

float EPSILON = .01;

int[] sources = new int[0];
int[] sinks = new int[sources.length];
int[] prev, next;

void setup() {
  size(w*2,h*2);
  setupGUI();
  resetValues();
  
  prev = new int[a.length];
  next = new int[a.length];
  //store period boundary for later
  for(int i=0;i<prev.length;++i) {
    prev[i] = (i-1+prev.length)%prev.length;
    next[i] = (i+1)%prev.length;
  }
  
  //aPrev[128][128] = a[128][128];
  //for(int i=0;i<sources.length;++i) {
  //  sources[i] = (int) random(a.length*a[0].length);
  //  sinks[i] = (int) random(a.length*a[0].length);
  //}
}

void resetValues() {
 for(int i=0;i<a.length;++i) {
    for(int j=0;j<a[0].length;++j) {
      //val[i][j] = sq(cos(PI*(i*i+j*j)/(2.0*sq(val.length))));
      
      //TURING
      //a[i][j] = 4.0 + random(-.04,.04);
      //b[i][j] = 4.0 + random(-.04,.04);
      //GRAY SCOTT
      a[i][j] = 1.0+random(-.01,.01);//
      b[i][j] = 0+random(.01);//
    }
  }
  
  //GRAY-SCOTT
  for(int i=w/2-10;i<w/2+10;++i) {
    for(int j=h/2-10;j<h/2+10;++j) {
      a[i][j] = .5+random(-.01,.01);
      b[i][j] = .25+random(-.01,.01);
    }
  }  
}

void draw() {
  background(255);
  //draw colors
  loadPixels();
  for(int i=0;i<width;++i) {
    for(int j=0;j<height;++j) {
      pixels[j*width+i] = color((b[i/2][j/2]-minB)/(maxB-minB)*255,0,(a[i/2][j/2]-minA)/(maxA-minA)*255);
    }
  }
  updatePixels();
  //for(int i=0;i<10;++i) stepDufortFrankel();
  for(int i=0;i<4;++i) stepSemiImplicit();
  println(frameCount + " " + maxA + " " + minA + " " + maxB + " " + minB);
}

void stepSemiImplicit() {
  //half step crank nicholson
  stepCrankNicholson();
  //half step explicit euler
  //stepEuler();
  stepRK4();
}

void stepCrankNicholson() {
  //simple diffusion with Red-Black Gauss-Seidel
  //Fn+1(i,j) = 1/(1+4r)*(r(sum(Fn(+-,+-)))+(1-4r)Fn(i,j)+rFn+1(+-,+-))
  float[][] stableA = new float[a.length][a.length];
  float[][] stableB = new float[a.length][a.length];
  //store the part of the implicit equation which does not change
  float diff = deltaT/(2.0*delxSq);
  for(int i = 0;i<w;++i) {
    for(int j= 0;j<h;++j) {
      //ISOTROPIC
      /*
      float diffA = aRate*deltaT/(2.0*delxSq);
      stableA[i][j] = (1.0-4.0*diffA)*a[i][j]+
        diffA*(a[prev[i]][j]+a[next[i]][j]+a[i][prev[j]]+a[i][next[j]]);
      float diffB = bRate*deltaT/(2.0*delxSq);
      stableB[i][j] = (1.0-4.0*diffB)*b[i][j]+
        diffB*(b[prev[i]][j]+b[next[i]][j]+b[i][prev[j]]+b[i][next[j]]);
      */
      //SIMPLE ANIISOTROPIC
      float DAx = .075+.05*i/w;
      float DAy = .1;
      float DBx = .05;
      float DBy = .05;
      stableA[i][j] = (1.0-2.0*DAx*diff-2*DAy*diff)*a[i][j]+
        diff*DAx*(a[prev[i]][j]+a[next[i]][j])+diff*DAy*(a[i][prev[j]]+a[i][next[j]]);
      stableB[i][j] = (1.0-2.0*DBx*diff-2.0*DBy*diff)*b[i][j]+
        diff*DBx*(b[prev[i]][j]+b[next[i]][j])+diff*DBy*(b[i][prev[j]]+b[i][next[j]]);
      }
  }

  //DO WE DO OVER-RELAXATION?
  float maxChange = 1;
  while(maxChange > EPSILON) {
    //red
    maxChange = 0;
    float pVal;
    
    //red
    for(int i = 0;i<w;++i) {
      for(int j= i%2;j<h;j+=2) {
        //ISOTROPIC
        /*
        //A
        pVal = a[i][j];
        float diffA = aRate*deltaT/(2.0*delxSq);
        a[i][j] = 1.0/(1+4.0*diffA)*(stableA[i][j]+diffA*(a[prev[i]][j]+a[next[i]][j]+a[i][prev[j]]+a[i][next[j]]));
        if(a[i][j]-pVal>maxChange) maxChange = a[i][j]-pVal;
        //B
        pVal = b[i][j];
        float diffB = bRate*deltaT/(2.0*delxSq);
        b[i][j] = 1.0/(1+4.0*diffB)*(stableB[i][j]+diffB*(b[prev[i]][j]+b[next[i]][j]+b[i][prev[j]]+b[i][next[j]]));
        if(b[i][j]-pVal>maxChange) maxChange = b[i][j]-pVal;
        */
        //SIMPLE ANISOTROPIC
        //A
        float DAx = .075+.05*i/w;
        float DAy = .1;
        float DBx = .05;
        float DBy = .05;
        pVal = a[i][j];
        a[i][j] = 1.0/(1+2.0*diff*(DAx+DAy))*(stableA[i][j]+DAx*diff*(a[prev[i]][j]+a[next[i]][j])+DAy*diff*(a[i][prev[j]]+a[i][next[j]]));
        if(a[i][j]-pVal>maxChange) maxChange = a[i][j]-pVal;
        //B
        pVal = b[i][j];
        b[i][j] = 1.0/(1+2.0*diff*(DBx+DBy))*(stableB[i][j]+DBx*diff*(b[prev[i]][j]+b[next[i]][j])+DBy*diff*(b[i][prev[j]]+b[i][next[j]]));
        if(b[i][j]-pVal>maxChange) maxChange = b[i][j]-pVal;
      }
    }
    
    //black
    for(int i = 0;i<w;++i) {
      for(int j= 1-i%2;j<h;j+=2) {
        //A
        pVal = a[i][j];
        float diffA = aRate*deltaT/(2.0*delxSq);
        a[i][j] = 1.0/(1+4.0*diffA)*(stableA[i][j]+diffA*(a[prev[i]][j]+a[next[i]][j]+a[i][prev[j]]+a[i][next[j]]));
        if(abs(a[i][j]-pVal)>maxChange) maxChange = abs(a[i][j]-pVal);
        //B
        pVal = b[i][j];
        float diffB = bRate*deltaT/(2.0*delxSq);
        b[i][j] = 1.0/(1+4.0*diffB)*(stableB[i][j]+diffB*(b[prev[i]][j]+b[next[i]][j]+b[i][prev[j]]+b[i][next[j]]));
        if(abs(b[i][j]-pVal)>maxChange) maxChange = abs(b[i][j]-pVal);
      }
    }
  }
}

void stepEuler() {
  maxA = 0;
  minA = 999;
  maxB = 0;
  minB = 999;
  for(int i=0;i<w;++i) {
    for(int j=0;j<h;++j) {
      float reactA = reactionAGS(a[i][j],b[i][j],i,j);
      float reactB = reactionBGS(a[i][j],b[i][j],i,j);
      a[i][j] += reactA*deltaT;
      b[i][j] += reactB*deltaT;
      if(a[i][j] < 0) a[i][j] = 0;
      if(b[i][j] < 0) b[i][j] = 0;
      if(a[i][j] > 1) a[i][j] = 1;
      if(b[i][j] > 1) b[i][j] = 1;
      if(a[i][j] > maxA) maxA = a[i][j];
      if(a[i][j] < minA) minA = a[i][j];
      if(b[i][j] > maxB) maxB = b[i][j];
      if(b[i][j] < minB) minB = b[i][j];
    }
  }
}

void stepRK4() {
  maxA = 0;
  minA = 999;
  maxB = 0;
  minB = 999;
  for(int i=0;i<w;++i) {
    for(int j=0;j<h;++j) {
      
      float k1A = deltaT*reactionAGS(a[i][j],b[i][j],i,j);
      float k1B = deltaT*reactionBGS(a[i][j],b[i][j],i,j);
      float k2A = deltaT*reactionAGS(a[i][j]+.5*k1A,b[i][j]+.5*k1B,i,j);
      float k2B = deltaT*reactionBGS(a[i][j]+.5*k1A,b[i][j]+.5*k1B,i,j);
      float k3A = deltaT*reactionAGS(a[i][j]+.5*k2A,b[i][j]+.5*k2B,i,j);
      float k3B = deltaT*reactionBGS(a[i][j]+.5*k2A,b[i][j]+.5*k2B,i,j);
      float k4A = deltaT*reactionAGS(a[i][j]+k3A,b[i][j]+k3B,i,j);
      float k4B = deltaT*reactionBGS(a[i][j]+k3A,b[i][j]+k3B,i,j);
      /*
      float k1A = deltaT*reactionATuring(a[i][j],b[i][j],i,j);
      float k1B = deltaT*reactionBTuring(a[i][j],b[i][j],i,j);
      float k2A = deltaT*reactionATuring(a[i][j]+.5*k1A,b[i][j]+.5*k1B,i,j);
      float k2B = deltaT*reactionBTuring(a[i][j]+.5*k1A,b[i][j]+.5*k1B,i,j);
      float k3A = deltaT*reactionATuring(a[i][j]+.5*k2A,b[i][j]+.5*k2B,i,j);
      float k3B = deltaT*reactionBTuring(a[i][j]+.5*k2A,b[i][j]+.5*k2B,i,j);
      float k4A = deltaT*reactionATuring(a[i][j]+k3A,b[i][j]+k3B,i,j);
      float k4B = deltaT*reactionBTuring(a[i][j]+k3A,b[i][j]+k3B,i,j);
      */
      
      a[i][j] += 1.0/6.0*(k1A+2*k2A+2*k3A+k4A);
      b[i][j] += 1.0/6.0*(k1B+2*k2B+2*k3B+k4B);
      if(a[i][j] < 0) a[i][j] = 0;
      if(b[i][j] < 0) b[i][j] = 0;
      if(a[i][j] > 1) a[i][j] = 1;
      if(b[i][j] > 1) b[i][j] = 1;
      if(a[i][j] > maxA) maxA = a[i][j];
      if(a[i][j] < minA) minA = a[i][j];
      if(b[i][j] > maxB) maxB = b[i][j];
      if(b[i][j] < minB) minB = b[i][j];
    }
  }
}

//REACTIONS

//GREY-SCOTT
float reactionAGS(float aVal, float bVal, int i, int j) {
  return reactionRate*(-aVal*bVal*bVal+F*(1-aVal));
}

float reactionBGS(float aVal, float bVal, int i, int j) {
  return reactionRate*(aVal*bVal*bVal-(F+k)*bVal);
}

//TURING
float reactionATuring(float aVal, float bVal, int i, int j) {
  return reactionRate*(aVal*bVal-aVal-decayA);
}

float reactionBTuring(float aVal, float bVal, int i, int j) {
  return reactionRate*(decayB-aVal*bVal);
}

//BRUSSELATOR
float reactionABrussel(float aVal, float bVal, int i, int j) {
  return reactionRate*(decayA-(1+decayB)*aVal+aVal*aVal*bVal);
}

float reactionBBrussel(float aVal, float bVal, int i, int j) {
  return reactionRate*(decayB*aVal-aVal*aVal*bVal);
}

//Gierer-Meinhardt
float reactionAGM(float aVal, float bVal, int i, int j) {
  return reactionRate*(gamma+aVal+aVal*aVal/bVal-decayA*aVal);
}

float reactionBGM(float aVal, float bVal, int i, int j) {
  return reactionRate*(aVal*aVal-decayB*bVal);
}

void keyPressed() {
  if(key == ' ') {
    resetValues();
  }
}
